"""f90nml
   ======

   A Fortran 90 namelist parser and generator.

   :copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""
from f90nml.parser import Parser
from f90nml.namelist import Namelist

__version__ = '0.17'


def read(nml_path, row_major=None, strict_logical=None):
    """Parse a Fortran 90 namelist file with file path ``nml_path``  return its
    contents as a ``Namelist``.

    >>> nml = f90nml.read('data.nml')

    This function is equivalent to the ``read`` function of the ``Parser``
    object.

    >>> parser = f90nml.Parser()
    >>> nml = parser.read('data.nml')

    Multidimensional array data contiguity is preserved by default, so that
    column-major Fortran data is represented as row-major Python list of
    lists.

    The ``row_major`` flag will reorder the data to preserve the index rules
    between Fortran to Python, but the data will be converted to row-major form
    (with respect to Fortran).

    The ``strict_logical`` flag will limit the parsing of non-delimited logical
    strings as logical values.  The default value is ``True``.

    When ``strict_logical`` is enabled, only ``.true.``, ``.t.``, ``true``, and
    ``t`` are interpreted as ``True``, and only ``.false.``, ``.f.``,
    ``false``, and ``.false.`` are interpreted as false.

    When ``strict_logical`` is disabled, any value starting with ``.t`` or
    ``t`` are interpreted as ``True``, while any string starting with ``.f`` or
    ``f`` is interpreted as ``False``."""

    parser = Parser()
    parser.row_major = row_major
    parser.strict_logical = strict_logical

    return parser.read(nml_path)


def write(nml, nml_path, force=False):
    """Output namelist ``nml`` to a Fortran 90 namelist file with file path
    ``nml_path``.

    >>> f90nml.write(nml, 'data.nml')

    This function is equivalent to the ``write`` function of the ``Namelist``
    object ``nml``.

    >>> nml.write('data.nml')

    By default, ``write`` will not overwrite an existing file.  To override
    this, use the ``force`` flag.

    >>> nml.write('data.nml', force=True)"""

    # Promote dicts to Namelists
    if not isinstance(nml, Namelist) and isinstance(nml, dict):
        nml_in = Namelist(nml)
    else:
        nml_in = nml

    nml_in.write(nml_path, force=force)


def patch(nml_path, nml_patch, out_path=None, row_major=None,
          strict_logical=None):
    """Create a new namelist based on an input namelist and reference dict.

    >>> f90nml.patch('data.nml', nml_patch, 'patched_data.nml')

    This function is equivalent to the ``read`` function of the ``Parser``
    object with the patch output arguments.

    >>> parser = f90nml.Parser()
    >>> nml = parser.read('data.nml', nml_path, 'patched_data.nml')

    A patched namelist file will retain any formatting or comments from the
    original namelist file.  Any modified values will be formatted based on the
    settings of the ``Namelist`` object."""

    parser = Parser()
    parser.row_major = row_major
    parser.strict_logical = strict_logical

    return parser.read(nml_path, nml_patch, out_path)



"""f90nml.parser
   =============

   Fortran namelist parser and tokenizer to convert contents into a hierarchy
   of dicts containing intrinsic Python data types.

   :copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""
import copy
import shlex
from string import whitespace

from f90nml.fpy import pyfloat, pycomplex, pybool, pystr
from f90nml.namelist import Namelist
from f90nml.findex import FIndex


class Parser(object):
    """shlex-based Fortran namelist parser."""

    def __init__(self):

        # Token management
        self.tokens = None
        self.token = None
        self.prior_token = None

        # Patching
        self.pfile = None

        # Control flags
        self._row_major = False
        self._strict_logical = True

        # Configuration
        self.comment_tokens = '!'

    @property
    def row_major(self):
        """Return true if multidimensional arrays are in row-major format."""
        return self._row_major

    @row_major.setter
    def row_major(self, value):
        """Validate and set row-major format for multidimensional arrays."""

        if value is not None:
            if not isinstance(value, bool):
                raise ValueError(
                    'f90nml: error: row_major must be a logical value.')
            else:
                self._row_major = value

    @property
    def strict_logical(self):
        """Return true for strict logical value parsing."""
        return self._strict_logical

    @strict_logical.setter
    def strict_logical(self, value):
        """Validate and set the strict logical flag."""

        if value is not None:
            if not isinstance(value, bool):
                raise ValueError(
                    'f90nml: error: strict_logical must be a logical value.')
            else:
                self._strict_logical = value

    def read(self, nml_fname, nml_patch_in=None, patch_fname=None):
        """Parse a Fortran 90 namelist file and store the contents.

        >>> from f90nml.parser import Parser
        >>> parser = Parser()
        >>> data_nml = parser.read('data.nml')"""

        nml_file = open(nml_fname, 'r')

        if nml_patch_in:
            if not isinstance(nml_patch_in, dict):
                nml_file.close()
                raise ValueError('Input patch must be a dict or an Namelist.')

            nml_patch = copy.deepcopy(Namelist(nml_patch_in))

            if not patch_fname:
                patch_fname = nml_fname + '~'
            elif nml_fname == patch_fname:
                nml_file.close()
                raise ValueError('f90nml: error: Patch filepath cannot be the '
                                 'same as the original filepath.')
            self.pfile = open(patch_fname, 'w')
        else:
            nml_patch = Namelist()

        f90lex = shlex.shlex(nml_file)
        f90lex.whitespace = ''
        f90lex.wordchars += '.-+'       # Include floating point tokens
        if nml_patch:
            f90lex.commenters = ''
        else:
            f90lex.commenters = self.comment_tokens

        self.tokens = iter(f90lex)

        nmls = Namelist()

        # TODO: Replace "while True" with an update_token() iterator
        self.update_tokens(write_token=False)
        while True:
            try:
                # Check for classic group terminator
                if self.token == 'end':
                    self.update_tokens()

                # Ignore tokens outside of namelist groups
                while self.token not in ('&', '$'):
                    self.update_tokens()

            except StopIteration:
                break

            # Create the next namelist
            self.update_tokens()
            g_name = self.token

            g_vars = Namelist()
            v_name = None

            grp_patch = nml_patch.get(g_name, {})

            # Populate the namelist group
            while g_name:

                if self.token not in ('=', '%', '('):
                    self.update_tokens()

                # Set the next active variable
                if self.token in ('=', '(', '%'):

                    try:
                        v_name, v_values = self.parse_variable(
                            g_vars, patch_nml=grp_patch)
                    except ValueError:
                        nml_file.close()
                        if self.pfile:
                            self.pfile.close()
                        raise

                    if v_name in g_vars:
                        v_prior_values = g_vars[v_name]
                        v_values = merge_values(v_prior_values, v_values)

                    if v_name in g_vars and isinstance(g_vars[v_name], dict):
                        g_vars[v_name].update(v_values)
                    else:
                        g_vars[v_name] = v_values

                    # Deselect variable
                    v_name = None
                    v_values = []

                # Finalise namelist group
                if self.token in ('/', '&', '$'):

                    # Append any remaining patched variables
                    for v_name, v_val in grp_patch.items():
                        g_vars[v_name] = v_val
                        v_strs = nmls.var_strings(v_name, v_val)
                        for v_str in v_strs:
                            self.pfile.write('    {0}\n'.format(v_str))

                    # Append the grouplist to the namelist
                    if g_name in nmls:
                        g_update = nmls[g_name]

                        # Update to list of groups
                        if not isinstance(g_update, list):
                            g_update = [g_update]

                        g_update.append(g_vars)

                    else:
                        g_update = g_vars

                    nmls[g_name] = g_update

                    # Reset state
                    g_name, g_vars = None, None

            try:
                self.update_tokens()
            except StopIteration:
                break

        nml_file.close()
        if self.pfile:
            self.pfile.close()

        return nmls

    def parse_variable(self, parent, patch_nml=None):
        """Parse a variable and return its name and values."""

        if not patch_nml:
            patch_nml = Namelist()

        v_name = self.prior_token
        v_values = []

        # Patch state
        patch_values = None
        write_token = v_name not in patch_nml

        if self.token == '(':

            v_idx_bounds = self.parse_indices()
            v_idx = FIndex(v_idx_bounds)

            self.update_tokens()
        else:
            v_idx = None

        if self.token == '%':

            # Resolve the derived type

            if parent and v_name in parent:
                v_parent = parent[v_name]
            else:
                v_parent = []

            self.update_tokens()
            self.update_tokens()

            v_att, v_att_vals = self.parse_variable(v_parent)

            next_value = Namelist()
            next_value[v_att] = v_att_vals
            self.append_value(v_values, next_value, v_idx)

        else:
            # Construct the variable array

            assert self.token == '='
            n_vals = None
            prior_ws_sep = ws_sep = False

            self.update_tokens()

            # Check if value is in the namelist patch
            if v_name in patch_nml:
                patch_values = patch_nml.f90repr(patch_nml.pop(v_name))
                if not isinstance(patch_values, list):
                    patch_values = [patch_values]

                for p_val in patch_values:
                    self.pfile.write(p_val)

            # Add variables until next variable trigger
            while (self.token not in ('=', '(', '%') or
                   (self.prior_token, self.token) == ('=', '(')):

                # Check for repeated values
                if self.token == '*':
                    n_vals = self.parse_value(write_token)
                    assert isinstance(n_vals, int)
                    self.update_tokens(write_token)
                elif not n_vals:
                    n_vals = 1

                # First check for implicit null values
                if self.prior_token in ('=', '%', ','):
                    if (self.token in (',', '/', '&', '$') and
                            not (self.prior_token == ',' and
                                 self.token in ('/', '&', '$'))):
                        self.append_value(v_values, None, v_idx, n_vals)

                elif self.prior_token == '*':

                    if self.token not in ('/', '&', '$'):
                        self.update_tokens(write_token)

                    if (self.token == '=' or (self.token in ('/', '&', '$') and
                                              self.prior_token == '*')):
                        next_value = None
                    else:
                        next_value = self.parse_value(write_token)

                    self.append_value(v_values, next_value, v_idx, n_vals)

                else:
                    next_value = self.parse_value(write_token)

                    # Finished reading old value, we can again write tokens
                    write_token = True

                    # Check for escaped strings
                    if (v_values and isinstance(v_values[-1], str) and
                            isinstance(next_value, str) and not prior_ws_sep):

                        quote_char = self.prior_token[0]
                        v_values[-1] = quote_char.join([v_values[-1],
                                                        next_value])
                    else:
                        self.append_value(v_values, next_value, v_idx, n_vals)

                # Exit for end of nml group (/, &, $) or null broadcast (=)
                if self.token in ('/', '&', '$', '='):
                    break
                else:
                    prior_ws_sep = ws_sep
                    ws_sep = self.update_tokens(write_token)

        if patch_values:
            v_values = patch_values

        if not v_idx:
            v_values = delist(v_values)

        return v_name, v_values

    def parse_indices(self):
        """Parse a sequence of Fortran vector indices as a list of tuples."""

        v_name = self.prior_token
        v_indices = []

        while self.token in (',', '('):
            v_indices.append(self.parse_index(v_name))

        return v_indices

    def parse_index(self, v_name):
        """Parse Fortran vector indices into a tuple of Python indices."""

        i_start = i_end = i_stride = None

        # Start index
        self.update_tokens()
        try:
            i_start = int(self.token)
            self.update_tokens()
        except ValueError:
            if self.token in (',', ')'):
                raise ValueError('{0} index cannot be empty.'.format(v_name))
            elif not self.token == ':':
                raise

        # End index
        if self.token == ':':
            self.update_tokens()
            try:
                i_end = 1 + int(self.token)
                self.update_tokens()
            except ValueError:
                if self.token == ':':
                    raise ValueError('{0} end index cannot be implicit '
                                     'when using stride.'.format(v_name))
                elif self.token not in (',', ')'):
                    raise
        elif self.token in (',', ')'):
            # Replace index with single-index range
            if i_start:
                i_end = 1 + i_start

        # Stride index
        if self.token == ':':
            self.update_tokens()
            try:
                i_stride = int(self.token)
            except ValueError:
                if self.token == ')':
                    raise ValueError('{0} stride index cannot be '
                                     'implicit.'.format(v_name))
                else:
                    raise

            if i_stride == 0:
                raise ValueError('{0} stride index cannot be zero.'
                                 ''.format(v_name))

            self.update_tokens()

        if self.token not in (',', ')'):
            raise ValueError('{0} index did not terminate '
                             'correctly.'.format(v_name))

        idx_triplet = (i_start, i_end, i_stride)
        return idx_triplet

    def parse_value(self, write_token=True):
        """Convert string repr of Fortran type to equivalent Python type."""
        v_str = self.prior_token

        # Construct the complex string
        if v_str == '(':
            v_re = self.token

            self.update_tokens(write_token)
            assert self.token == ','

            self.update_tokens(write_token)
            v_im = self.token

            self.update_tokens(write_token)
            assert self.token == ')'

            self.update_tokens(write_token)
            v_str = '({0}, {1})'.format(v_re, v_im)

        recast_funcs = [int, pyfloat, pycomplex, pybool, pystr]

        for f90type in recast_funcs:
            try:
                # Unclever hack.. integrate this better
                if f90type == pybool:
                    value = pybool(v_str, self.strict_logical)
                else:
                    value = f90type(v_str)
                return value
            except ValueError:
                continue

    def update_tokens(self, write_token=True):
        """Update tokens to the next available values."""

        ws_sep = False
        next_token = next(self.tokens)

        if self.pfile and write_token:
            self.pfile.write(self.token)

        # Commas between values are interpreted as whitespace
        if self.token == ',':
            ws_sep = True

        while next_token in tuple(whitespace + '!'):

            if self.pfile:
                if next_token == '!':
                    while not next_token == '\n':
                        self.pfile.write(next_token)
                        next_token = next(self.tokens)
                self.pfile.write(next_token)

            ws_sep = True
            next_token = next(self.tokens)

        self.token, self.prior_token = next_token, self.token

        return ws_sep

    def append_value(self, v_values, next_value, v_idx=None, n_vals=1):
        """Update a list of parsed values with a new value."""

        for _ in range(n_vals):
            if v_idx:
                v_i = next(v_idx)

                if not self.row_major:
                    v_i = v_i[::-1]

                # Multidimensional arrays
                # TODO: support both row and column ordering in Python

                v_tmp = v_values
                for idx in v_i[:-1]:
                    try:
                        v_tmp = v_tmp[idx - 1]
                    except IndexError:
                        size = len(v_tmp)
                        v_tmp.extend([] for i in range(size, idx))
                        v_tmp = v_tmp[idx - 1]

                try:
                    v_tmp[v_i[-1] - 1] = next_value
                except IndexError:
                    size = len(v_tmp)
                    v_tmp.extend(None for i in range(size, v_i[-1]))
                    v_tmp[v_i[-1] - 1] = next_value
            else:
                v_values.append(next_value)


# Support functions

def merge_values(src, new):
    """Merge two lists or dicts into a single element."""

    if isinstance(src, dict) and isinstance(new, dict):
        return merge_dicts(src, new)
    else:
        if not isinstance(src, list):
            src = [src]
        if not isinstance(new, list):
            new = [new]

        return merge_lists(src, new)


def merge_lists(src, new):
    """Update a value list with a list of new or updated values."""

    l_min, l_max = (src, new) if len(src) < len(new) else (new, src)

    l_min.extend(None for i in range(len(l_min), len(l_max)))

    for i, val in enumerate(new):
        if isinstance(val, dict) and isinstance(src[i], dict):
            new[i] = merge_dicts(src[i], val)
        elif isinstance(val, list) and isinstance(src[i], list):
            new[i] = merge_lists(src[i], val)
        elif val is not None:
            new[i] = val
        else:
            new[i] = src[i]

    return new


def merge_dicts(src, patch):
    """Merge contents of dict `patch` into `src`."""

    for key in patch:
        if key in src:
            if isinstance(src[key], dict) and isinstance(patch[key], dict):
                merge_dicts(src[key], patch[key])
            else:
                src[key] = merge_values(src[key], patch[key])
        else:
            src[key] = patch[key]

    return src


def delist(values):
    """Reduce lists of zero or one elements to individual values."""
    assert isinstance(values, list)

    if not values:
        return None
    elif len(values) == 1:
        return values[0]
    else:
        return values



"""f90nml.findex
   =============

   Column-major Fortran iterator of indices across multipe dimensions.

   :copyright: Copyright 2015 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""


class FIndex(object):
    """Column-major multidimensional index iterator"""

    def __init__(self, bounds):
        self.start = [1 if not b[0] else b[0] for b in bounds]
        self.end = [b[1] for b in bounds]
        self.step = [1 if not b[2] else b[2] for b in bounds]

        self.current = self.start[:]

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.end[-1] and self.current[-1] >= self.end[-1]:
            raise StopIteration

        state = self.current[:]
        # Allow the final index to exceed self.end[-1] as a finalisation check
        for rank, idx in enumerate(self.current):
            if ((not self.end[rank] or idx < (self.end[rank] - 1)) or
                    rank == (len(self.current) - 1)):
                self.current[rank] = idx + self.step[rank]
                break
            else:
                self.current[rank] = self.start[rank]

        return state



"""f90nml.fpy
   =============

   Module for conversion between basic data types and Fortran string
   representations.

   :copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""


def pyfloat(v_str):
    """Convert string repr of Fortran floating point to Python double."""
    # NOTE: There is no loss of information from SP to DP floats

    return float(v_str.lower().replace('d', 'e'))


def pycomplex(v_str):
    """Convert string repr of Fortran complex to Python complex."""
    assert isinstance(v_str, str)

    if v_str[0] == '(' and v_str[-1] == ')' and len(v_str.split(',')) == 2:
        v_re, v_im = v_str[1:-1].split(',', 1)

        # NOTE: Failed float(str) will raise ValueError
        return complex(pyfloat(v_re), pyfloat(v_im))
    else:
        raise ValueError('{0} must be in complex number form (x, y).'
                         ''.format(v_str))


def pybool(v_str, strict_logical=True):
    """Convert string repr of Fortran logical to Python logical."""
    assert isinstance(v_str, str)
    assert isinstance(strict_logical, bool)

    if strict_logical:
        v_bool = v_str.lower()
    else:
        try:
            if v_str.startswith('.'):
                v_bool = v_str[1].lower()
            else:
                v_bool = v_str[0].lower()
        except IndexError:
            raise ValueError('{0} is not a valid logical constant.'
                             ''.format(v_str))

    if v_bool in ('.true.', '.t.', 'true', 't'):
        return True
    elif v_bool in ('.false.', '.f.', 'false', 'f'):
        return False
    else:
        raise ValueError('{0} is not a valid logical constant.'.format(v_str))


def pystr(v_str):
    """Convert string repr of Fortran string to Python string."""
    assert isinstance(v_str, str)

    if v_str[0] in ("'", '"') and v_str[0] == v_str[-1]:
        return v_str[1:-1]
    else:
        return v_str



"""f90nml.namelist
   ===============

   Tools for creating Fortran namelist files from Python ``dict``s.

   :copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""
from __future__ import print_function

import os
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class Namelist(OrderedDict):
    """Case-insensitive Python dict"""

    def __init__(self, *args, **kwds):

        # If using (unordered) dict, then resort the keys for reproducibility
        s_args = list(args)
        if (args and not isinstance(args[0], OrderedDict) and
                isinstance(args[0], dict)):
            s_args[0] = sorted(args[0].items())

        super(Namelist, self).__init__(*s_args, **kwds)

        # Convert any internal dicts to Namelists
        for key, val in self.items():
            if isinstance(val, dict):
                self[key] = Namelist(val)

        # Formatting properties
        self._colwidth = 72
        self._indent = 4 * ' '
        self._end_comma = False
        self._uppercase = False
        self._floatformat = ''
        self._logical_repr = {False: '.false.', True: '.true.'}

        # Representatation functions
        self.f90str = {
            bool:
                lambda x: self.logical_repr[x],
            int:
                lambda x: str(x),
            float:
                lambda x: '{0:{fmt}}'.format(x, fmt=self.floatformat),
            complex:
                lambda x: '({0}, {1})'.format(x.real, x.imag),
            str:
                lambda x: repr(x).replace("\\'", "''").replace('\\"', '""'),
            type(None):
                lambda x: ''
        }

    def __contains__(self, key):
        return super(Namelist, self).__contains__(key.lower())

    def __delitem__(self, key):
        return super(Namelist, self).__delitem__(key.lower())

    def __getitem__(self, key):
        return super(Namelist, self).__getitem__(key.lower())

    def __setitem__(self, key, value):
        super(Namelist, self).__setitem__(key.lower(), value)

    # Format configuration

    # Column width
    @property
    def colwidth(self):
        """Return the target column width of the namelist file."""
        return self._colwidth

    @colwidth.setter
    def colwidth(self, width):
        """Validate and set the column width."""
        if isinstance(width, int):
            if width >= 0:
                self._colwidth = width
            else:
                raise ValueError('Column width must be nonnegative.')
        else:
            raise TypeError('Column width must be a nonnegative integer.')

    # Variable indent
    @property
    def indent(self):
        """Return the indentation string within namelist group entries."""
        return self._indent

    @indent.setter
    def indent(self, value):
        """Validate and set the indent width, either as an explicit whitespace
        string or by the number of whitespace characters.
        """

        # Explicit indent setting
        if isinstance(value, str):
            if value.isspace():
                self._indent = value
            else:
                raise ValueError('String indentation can only contain '
                                 'whitespace.')

        # Set indent width
        elif isinstance(value, int):
            if value >= 0:
                self._indent = value * ' '
            else:
                raise ValueError('Indentation spacing must be nonnegative.')

        else:
            raise TypeError('Indentation must be specified by string or space '
                            'width.')

    # Terminal comma
    @property
    def end_comma(self):
        """Return True if entries are terminated with commas."""
        return self._end_comma

    @end_comma.setter
    def end_comma(self, value):
        """Validate and set the comma termination flag."""
        if not isinstance(value, bool):
            raise TypeError('end_comma attribute must be a logical type.')
        self._end_comma = value

    # Uppercase
    @property
    def uppercase(self):
        """Return True if names are displayed in upper case."""
        return self._uppercase

    @uppercase.setter
    def uppercase(self, value):
        """Validate and set the upper case flag."""
        if not isinstance(value, bool):
            raise TypeError('uppercase attribute must be a logical type.')
        self._uppercase = value

    # Float format
    @property
    def floatformat(self):
        """Return the current floating point format code."""
        return self._floatformat

    @floatformat.setter
    def floatformat(self, value):
        """Validate and set the upper case flag."""
        if isinstance(value, str):
            # Duck-test the format string; raise ValueError on fail
            '{0:{1}}'.format(1.23, value)

            self._floatformat = value
        else:
            raise TypeError('Floating point format code must be a string.')

    # Logical representation
    # NOTE: This presumes that bools and ints are identical as dict keys
    @property
    def logical_repr(self):
        """Return the namelist representations of logical values."""
        return self._logical_repr

    @logical_repr.setter
    def logical_repr(self, value):
        """Set the namelist representations of logical values."""

        if not any(isinstance(value, t) for t in (list, tuple)):
            raise TypeError("Logical representation must be a tuple with "
                            "a valid true and false value.")
        if not len(value) == 2:
            raise ValueError("List must contain two values.")

        self.false_repr = value[0]
        self.true_repr = value[1]

    @property
    def true_repr(self):
        """Return the namelist representation of logical true."""
        return self._logical_repr[1]

    @true_repr.setter
    def true_repr(self, value):
        """Validate and set the logical true representation."""
        if isinstance(value, str):
            if not (value.lower().startswith('t') or
                    value.lower().startswith('.t')):
                raise ValueError("Logical true representation must start with "
                                 "'T' or '.T'.")
            else:
                self._logical_repr[1] = value
        else:
            raise TypeError('Logical true representation must be a string.')

    @property
    def false_repr(self):
        """Return the namelist representation of logical false."""
        return self._logical_repr[0]

    @false_repr.setter
    def false_repr(self, value):
        """Validate and set the logical false representation."""
        if isinstance(value, str):
            if not (value.lower().startswith('f') or
                    value.lower().startswith('.f')):
                raise ValueError("Logical false representation must start "
                                 "with 'F' or '.F'.")
            else:
                self._logical_repr[0] = value
        else:
            raise TypeError('Logical false representation must be a string.')

    # File output

    def write(self, nml_path, force=False):
        """Output dict to a Fortran 90 namelist file."""

        if not force and os.path.isfile(nml_path):
            raise IOError('File {0} already exists.'.format(nml_path))

        with open(nml_path, 'w') as nml_file:
            for grp_name, grp_vars in self.items():
                # Check for repeated namelist records (saved as lists)
                if isinstance(grp_vars, list):
                    for g_vars in grp_vars:
                        self.write_nmlgrp(grp_name, g_vars, nml_file)
                else:
                    self.write_nmlgrp(grp_name, grp_vars, nml_file)

        if self.items():
            with open(nml_path, 'rb+') as nml_file:
                nml_file.seek(-1, os.SEEK_END)
                nml_file.truncate()

    def write_nmlgrp(self, grp_name, grp_vars, nml_file):
        """Write namelist group to target file."""

        if self.uppercase:
            grp_name = grp_name.upper()

        print('&{0}'.format(grp_name), file=nml_file)

        for v_name, v_val in grp_vars.items():

            for v_str in self.var_strings(v_name, v_val):
                nml_line = self.indent + '{0}'.format(v_str)
                print(nml_line, file=nml_file)

        print('/', file=nml_file)
        print(file=nml_file)

    def var_strings(self, v_name, v_values, v_idx=None):
        """Convert namelist variable to list of fixed-width strings."""

        if self.uppercase:
            v_name = v_name.upper()

        var_strs = []

        # Parse derived type contents
        if isinstance(v_values, dict):
            for f_name, f_vals in v_values.items():
                v_title = '%'.join([v_name, f_name])

                v_strs = self.var_strings(v_title, f_vals)
                var_strs.extend(v_strs)

        # Parse an array of derived types
        elif (isinstance(v_values, list) and
              any(isinstance(v, dict) for v in v_values) and
              all((isinstance(v, dict) or v is None) for v in v_values)):
            for idx, val in enumerate(v_values, start=1):

                if val is None:
                    continue

                v_title = v_name + '({0})'.format(idx)

                v_strs = self.var_strings(v_title, val)
                var_strs.extend(v_strs)

        # Parse a multidimensional array
        # TODO: Merge with the array of derived types
        elif (isinstance(v_values, list) and
              any(isinstance(v, list) for v in v_values) and
              all((isinstance(v, list) or v is None) for v in v_values)):

            if not v_idx:
                v_idx = []

            v_title = v_name
            for idx, val in enumerate(v_values, start=1):

                v_idx_new = v_idx + [idx]
                v_strs = self.var_strings(v_title, val, v_idx_new)
                var_strs.extend(v_strs)

        else:
            if not isinstance(v_values, list):
                v_values = [v_values]

            if v_idx:
                v_name += '(:, ' + ', '.join(str(i) for i in v_idx[::-1]) + ')'

            # Split output across multiple lines (if necessary)
            val_strs = []

            val_line = ''
            for v_val in v_values:

                v_width = self.colwidth - len(self.indent + v_name + ' = ')

                if len(val_line) < v_width:
                    val_line += self.f90repr(v_val) + ', '

                if len(val_line) >= v_width:
                    val_strs.append(val_line.rstrip())
                    val_line = ''

            # Append any remaining values
            if val_line:
                if (self.end_comma or v_values[-1] is None):
                    val_strs.append(val_line)
                else:
                    val_strs.append(val_line[:-2])

            # Complete the set of values
            var_strs.append('{0} = {1}'.format(v_name, val_strs[0]).strip())

            for v_str in val_strs[1:]:
                var_strs.append(' ' * (len(v_name + ' = ')) + v_str)

        return var_strs

    def f90repr(self, value):
        """Convert primitive Python types to equivalent Fortran strings."""

        try:
            return self.f90str[type(value)](value)
        except KeyError:
            raise ValueError('Type {0} of {1} cannot be converted to a '
                             'Fortran type.'.format(type(value), value))



import os
import sys
import unittest

sys.path.insert(1, '../')
import f90nml
from f90nml.fpy import pybool
from f90nml.namelist import Namelist
from f90nml.findex import FIndex


class Test(unittest.TestCase):

    def setUp(self):
        self.empty_nml = {'empty_nml': {}}

        self.null_nml = {
                'null_nml': {'null_value': None},
                'null_comma_nml': {'null_comma': None},
                'null_nocomma_rpt_nml': {
                    'null_one': None,
                    'null_two': None,
                    }
                }

        self.unset_nml = {
                'unset_nml': {
                    'x': None,
                    'y': None
                    }
                }

        self.types_nml = {
                'types_nml': {
                    'v_integer': 1,
                    'v_float': 1.0,
                    'v_complex': 1+2j,
                    'v_logical': True,
                    'v_string': 'Hello',
                    }
                }

        self.vector_nml = {
                'vector_nml': {
                    'v': [1, 2, 3, 4, 5],
                    'v_idx': [1, 2, 3, 4],
                    'v_idx_ooo': [1, 2, 3, 4],
                    'v_range': [1, 2, 3, 4],
                    'v_implicit_start': [1, 2, 3, 4],
                    'v_implicit_end': [1, 2, 3, 4],
                    'v_implicit_all': [1, 2, 3, 4],
                    'v_null_start': [None, 2, 3, 4],
                    'v_null_interior': [1, 2, None, 4],
                    'v_null_end': [1, 2, 3, None],
                    'v_zero': [1, 0, 3],
                    'v_stride': [1, None, 3, None, 5, None, 7],
                    'v_single': [1],
                    'v_implicit_merge': [1, 2],
                    'v_explicit_merge': [1, 2],
                    }
                }

        self.multidim_nml = {
                'multidim_nml': {
                    'v2d': [[1, 2], [3, 4]],
                    'v3d': [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
                    'w3d': [[[1, 2, 3, 4],
                             [5, 6, 7, 8],
                             [9, 10, 11, 12]],
                            [[13, 14, 15, 16],
                             [17, 18, 19, 20],
                             [21, 22, 23, 24]]],
                    'v2d_explicit': [[1, 2], [3, 4]],
                    'v2d_outer': [[1], [2], [3], [4]],
                    'v2d_inner': [[1, 2, 3, 4]],
                    }
                }

        self.md_rowmaj_nml = {
                'multidim_nml': {
                    'v2d': [[1, 3], [2, 4]],
                    'v3d': [[[1, 5], [3, 7]], [[2, 6], [4, 8]]],
                    'w3d': [[[1, 13], [5, 17], [9, 21]],
                            [[2, 14], [6, 18], [10, 22]],
                            [[3, 15], [7, 19], [11, 23]],
                            [[4, 16], [8, 20], [12, 24]]],
                    'v2d_explicit': [[1, 3], [2, 4]],
                    'v2d_outer': [[1, 2, 3, 4]],
                    'v2d_inner': [[1], [2], [3], [4]],
                    }
                }

        self.float_nml = {
                'float_nml': {
                    'v_float': 1.,
                    'v_decimal_end': 1.,
                    'v_negative': -1.,
                    'v_single': 1.,
                    'v_double': 1.,
                    'v_single_upper': 1.,
                    'v_double_upper': 1.,
                    'v_positive_index': 10.,
                    'v_negative_index': 0.1,
                    }
                }

        self.string_nml = {
                'string_nml': {
                    'str_basic': 'hello',
                    'str_no_delim': 'hello',
                    'str_no_delim_no_esc': "a''b",
                    'single_esc_delim': "a 'single' delimiter",
                    'double_esc_delim': 'a "double" delimiter',
                    'double_nested': "''x'' \"y\"",
                    'str_list': ['a', 'b', 'c'],
                    'slist_no_space': ['a', 'b', 'c'],
                    'slist_no_quote': ['a', 'b', 'c'],
                    }
                }

        self.dtype_nml = {
                'dtype_nml': {
                    'dt_scalar': {'val': 1},
                    'dt_stack': {'outer': {'inner': 2}},
                    'dt_vector': {'vec': [1, 2, 3]}
                    },
                'dtype_multi_nml': {
                    'dt': {
                        'x': 1,
                        'y': 2,
                        'z': 3,
                        }
                    },
                'dtype_nested_nml': {
                    'f': {
                        'g': {
                            'x': 1,
                            'y': 2,
                            'z': 3,
                            }
                        }
                    },
                'dtype_field_idx_nml': {
                    'f': {
                        'x': [1, 2, 3]}
                    },
                'dtype_vec_nml': {
                    'a': {
                        'b': [
                            {'c': 1, 'd': 2},
                            {'c': 3, 'd': 4},
                            {'c': 5, 'd': 6}
                            ]
                        }
                    },
                'dtype_sparse_vec_nml': {
                    'a': {
                        'b': [None, {'c': 2}]
                        }
                    },
                'dtype_single_value_vec_nml': {
                    'a': [{'b': 1}]
                    },
                'dtype_single_vec_merge_nml': {
                    'a': {
                        'b': [{'c': 1, 'd': 2}]
                        }
                    }
                }

        self.bcast_nml = {
                'bcast_nml': {
                    'x': [2.0, 2.0],
                    'y': [None, None, None],
                    'z': [True, True, True, True],
                    },
                'bcast_endnull_nml': {
                    'x': [2.0, 2.0],
                    'y': [None, None, None],
                    }
                }

        self.comment_nml = {
                'comment_nml': {
                    'v_cmt_inline': 123,
                    'v_cmt_in_str': 'This token ! is not a comment',
                    'v_cmt_after_str': 'This ! is not a comment',
                    }
                }

        self.comment_alt_nml = {
                'comment_alt_nml': {
                    'x': 1,
                    'z': 3}
                }

        self.grp_repeat_nml = {
                'grp_repeat_nml': [{'x': 1}, {'x': 2}],
                'case_check_nml': [{'y': 1}, {'y': 2}],
                }

        self.f77_nml = {
                'f77_nml': {'x': 123},
                'next_f77_nml': {'y': 'abc'},
                }

        self.dollar_nml = {'dollar_nml': {'v': 1.}}

        self.multiline_nml = {
                'multiline_nml': {
                    'x': [
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                        1, 1, 1, 1, 1, 1, 1, 1
                        ]
                    }
                }

        self.ext_token_nml = {'ext_token_nml': {'x': 1}}

    # Support functions
    def assert_file_equal(self, source_fname, target_fname):
        with open(source_fname) as source:
            with open(target_fname) as target:
                source_str = source.read()
                target_str = target.read()
                self.assertEqual(source_str, target_str)

    def assert_write(self, nml, target_fname):
        tmp_fname = 'tmp.nml'
        f90nml.write(nml, tmp_fname)
        try:
            self.assert_file_equal(tmp_fname, target_fname)
        finally:
            os.remove(tmp_fname)

    # Tests
    def test_empty(self):
        test_nml = f90nml.read('empty.nml')
        self.assertEqual(self.empty_nml, test_nml)
        self.assert_write(test_nml, 'empty.nml')

    def test_null(self):
        test_nml = f90nml.read('null.nml')
        self.assertEqual(self.null_nml, test_nml)
        self.assert_write(test_nml, 'null_target.nml')

    def test_unset(self):
        test_nml = f90nml.read('unset.nml')
        self.assertEqual(self.unset_nml, test_nml)
        self.assert_write(test_nml, 'unset.nml')

    def test_types(self):
        test_nml = f90nml.read('types.nml')
        self.assertEqual(self.types_nml, test_nml)
        self.assert_write(test_nml, 'types.nml')

    def test_vector(self):
        test_nml = f90nml.read('vector.nml')
        self.assertEqual(self.vector_nml, test_nml)
        self.assert_write(test_nml, 'vector_target.nml')

    def test_multidim(self):
        test_nml = f90nml.read('multidim.nml')
        self.assertEqual(self.multidim_nml, test_nml)
        self.assert_write(test_nml, 'multidim_target.nml')

    def test_rowmaj_multidim(self):
        test_nml = f90nml.read('multidim.nml', row_major=True)
        self.assertEqual(self.md_rowmaj_nml, test_nml)

    def test_flag_syntax(self):
        self.assertRaises(ValueError, f90nml.read, 'index_empty.nml',
                          row_major='abc')
        self.assertRaises(ValueError, f90nml.read, 'index_empty.nml',
                          strict_logical='abc')

    def test_float(self):
        test_nml = f90nml.read('float.nml')
        self.assertEqual(self.float_nml, test_nml)
        self.assert_write(test_nml, 'float_target.nml')

    def test_string(self):
        test_nml = f90nml.read('string.nml')
        self.assertEqual(self.string_nml, test_nml)
        self.assert_write(test_nml, 'string_target.nml')

    def test_dtype(self):
        test_nml = f90nml.read('dtype.nml')
        self.assertEqual(self.dtype_nml, test_nml)
        self.assert_write(test_nml, 'dtype_target.nml')

    def test_bcast(self):
        test_nml = f90nml.read('bcast.nml')
        self.assertEqual(self.bcast_nml, test_nml)
        self.assert_write(test_nml, 'bcast_target.nml')

    def test_comment(self):
        test_nml = f90nml.read('comment.nml')
        self.assertEqual(self.comment_nml, test_nml)
        self.assert_write(test_nml, 'comment_target.nml')

    def test_comment_alt(self):
        parser = f90nml.Parser()
        parser.comment_tokens = '#'
        test_nml = parser.read('comment_alt.nml')
        self.assertEqual(self.comment_alt_nml, test_nml)

    def test_grp_repeat(self):
        test_nml = f90nml.read('grp_repeat.nml')
        self.assertEqual(self.grp_repeat_nml, test_nml)
        self.assert_write(test_nml, 'grp_repeat_target.nml')

    def test_f77(self):
        test_nml = f90nml.read('f77.nml')
        self.assertEqual(self.f77_nml, test_nml)
        self.assert_write(test_nml, 'f77_target.nml')

    def test_dollar(self):
        test_nml = f90nml.read('dollar.nml')
        self.assertEqual(self.dollar_nml, test_nml)
        self.assert_write(test_nml, 'dollar_target.nml')

    def test_multiline(self):
        test_nml = f90nml.read('multiline.nml')
        self.assertEqual(self.multiline_nml, test_nml)
        self.assert_write(test_nml, 'multiline.nml')

    def test_ext_token(self):
        test_nml = f90nml.read('ext_token.nml')
        self.assertEqual(self.ext_token_nml, test_nml)

    def test_write_existing_file(self):
        tmp_fname = 'tmp.nml'
        open(tmp_fname, 'w').close()
        test_nml = f90nml.read('empty.nml')
        self.assertRaises(IOError, test_nml.write, tmp_fname)
        os.remove(tmp_fname)

    def test_pop_key(self):
        test_nml = f90nml.read('empty.nml')
        test_nml.pop('empty_nml')
        self.assertEqual(test_nml, f90nml.namelist.Namelist())

    def test_patch(self):
        patch_nml = f90nml.read('types_patch.nml')
        f90nml.patch('types.nml', patch_nml, 'tmp.nml')
        test_nml = f90nml.read('tmp.nml')
        try:
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('tmp.nml')

    def test_patch_valueerror(self):
        self.assertRaises(ValueError, f90nml.patch, 'types.nml', 'xyz',
                          'tmp.nml')

    def test_default_patch(self):
        patch_nml = f90nml.read('types_patch.nml')
        f90nml.patch('types.nml', patch_nml)
        test_nml = f90nml.read('types.nml~')
        try:
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('types.nml~')

    def test_no_selfpatch(self):
        patch_nml = f90nml.read('types_patch.nml')
        self.assertRaises(ValueError, f90nml.patch,
                          'types.nml', patch_nml, 'types.nml')

    def test_comment_patch(self):
        nml = {'comment_nml': {'v_cmt_inline': 456}}
        try:
            f90nml.patch('comment.nml', nml, 'tmp.nml')
            self.assert_file_equal('comment_patch.nml', 'tmp.nml')
        finally:
            os.remove('tmp.nml')

    def test_index_syntax(self):
        self.assertRaises(ValueError, f90nml.read, 'index_empty.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_start.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_empty_end.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_end.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_empty_stride.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_stride.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_zero_stride.nml')

    def test_f90repr(self):
        nml = Namelist()
        self.assertEqual(nml.f90repr(1), '1')
        self.assertEqual(nml.f90repr(1.), '1.0')
        self.assertEqual(nml.f90repr(1+2j), '(1.0, 2.0)')
        self.assertEqual(nml.f90repr(True), '.true.')
        self.assertEqual(nml.f90repr(False), '.false.')
        self.assertEqual(nml.f90repr('abc'), "'abc'")

        for ptype in ({}, [], set()):
            self.assertRaises(ValueError, nml.f90repr, ptype)

    def test_pybool(self):
        for fstr_true in ('true', '.true.', 't', '.t.'):
            self.assertEqual(pybool(fstr_true), True)

        for fstr_false in ('false', '.false.', 'f', '.f.'):
            self.assertEqual(pybool(fstr_false), False)

        for fstr_true in ('ture', '.t'):
            self.assertEqual(pybool(fstr_true, strict_logical=False), True)

        for fstr_false in ('flase', '.f'):
            self.assertEqual(pybool(fstr_false, strict_logical=False), False)

        for fstr in ('ture', '.t', 'flase', '.f'):
            self.assertRaises(ValueError, pybool, fstr)

        for fstr in ('g', '.', 'xyz'):
            self.assertRaises(ValueError, pybool, fstr, strict_logical=False)

    def test_close_patch_on_error(self):
        patch = {'tmp_nml': {'tmp_val': 0}}
        self.assertRaises(ValueError, f90nml.patch, 'index_empty.nml', patch,
                                                    'tmp.nml')
        os.remove('tmp.nml')

    def test_indent(self):
        test_nml = f90nml.read('types.nml')

        test_nml.indent = 2
        self.assert_write(test_nml, 'types_indent_2.nml')

        test_nml.indent = '\t'
        self.assert_write(test_nml, 'types_indent_tab.nml')

        self.assertRaises(ValueError, setattr, test_nml, 'indent', -4)
        self.assertRaises(ValueError, setattr, test_nml, 'indent', 'xyz')
        self.assertRaises(TypeError, setattr, test_nml, 'indent', [1, 2, 3])

    def test_colwidth(self):
        test_nml = f90nml.read('multiline.nml')
        test_nml.colwidth = 40
        self.assert_write(test_nml, 'multiline_colwidth.nml')

        self.assertRaises(ValueError, setattr, test_nml, 'colwidth', -1)
        self.assertRaises(TypeError, setattr, test_nml, 'colwidth', 'xyz')

    def test_end_comma(self):
        test_nml = f90nml.read('types.nml')
        test_nml.end_comma = True
        self.assert_write(test_nml, 'types_end_comma.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'end_comma', 'xyz')

    def test_uppercase(self):
        test_nml = f90nml.read('types.nml')
        test_nml.uppercase = True
        self.assert_write(test_nml, 'types_uppercase.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'uppercase', 'xyz')

    def test_floatformat(self):
        test_nml = f90nml.read('float.nml')
        test_nml.floatformat = '.3f'
        self.assert_write(test_nml, 'float_format.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'floatformat', 123)

    def test_logical_repr(self):
        test_nml = f90nml.read('logical.nml', strict_logical=False)
        test_nml.true_repr = 'T'
        test_nml.false_repr = 'F'

        self.assertEqual(test_nml.false_repr, test_nml.logical_repr[0])
        self.assertEqual(test_nml.true_repr, test_nml.logical_repr[1])
        self.assert_write(test_nml, 'logical_repr.nml')

        test_nml.logical_repr = 'F', 'T'
        self.assert_write(test_nml, 'logical_repr.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'true_repr', 123)
        self.assertRaises(TypeError, setattr, test_nml, 'false_repr', 123)
        self.assertRaises(ValueError, setattr, test_nml, 'true_repr', 'xyz')
        self.assertRaises(ValueError, setattr, test_nml, 'false_repr', 'xyz')
        self.assertRaises(TypeError, setattr, test_nml, 'logical_repr', 'xyz')
        self.assertRaises(ValueError, setattr, test_nml, 'logical_repr', [])

    def test_findex_iteration(self):
        rng = [(None, 5, None)]
        fidx = iter(FIndex(rng))

        for i, j in enumerate(fidx, start=1):
            self.assertEqual(i, j[0])

    def test_dict_write(self):
        self.assert_write(self.types_nml, 'types_dict.nml')


if __name__ == '__main__':
    if os.path.isfile('tmp.nml'):
        os.remove('tmp.nml')
    unittest.main()
